@@ -5,8 +5,7 @@
  * Copyright (c) 2002 Jamie Wilkinson
  *
  * This program is free software, you may copy and/or modify as per
- * the  GNU General Public License (version 2, or at your discretion,
- * any later version).  This is the same license as libmad.
+ * the  GNU General Public License version 2.
  *
  * The code in py_madfile_read was copied from the madlld program, which
  * can be found at http://www.bsd-dk.dk/~elrond/audio/madlld/ and carries
@@ -49,6 +48,9 @@
 #include <stdio.h>
 #include <stdint.h>
 #include <errno.h>
+#include <assert.h>
+#include <fcntl.h>
+#include <stdint.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/mman.h>
@@ -61,6 +63,19 @@
 #define PyObject_DEL(op) PyMem_DEL((op))
 #endif
 
+#ifndef PyVarObject_HEAD_INIT
+#define PyVarObject_HEAD_INIT(type, size) PyObject_HEAD_INIT(type) size,
+#endif
+
+#if PY_MAJOR_VERSION >= 3
+#define MADFILE_GETATTR 0
+#define PyInt_FromLong PyLong_FromLong
+#define PyInt_AsLong PyLong_AsLong
+#else
+#define MADFILE_GETATTR (getattrfunc) py_madfile_getattr
+#define PyBytes_AsStringAndSize PyString_AsStringAndSize
+#endif
+
 #define ERROR_MSG_SIZE 512
 #define MAD_BUF_SIZE                                                           \
   (5 * 8192) /* should be a multiple of 4 >= 4096 and large enough to capture  \
@@ -69,41 +84,93 @@
 
 /* local helpers */
 static unsigned long calc_total_time(PyObject *);
-static signed short int madfixed_to_short(mad_fixed_t);
+static int16_t madfixed_to_int16(mad_fixed_t);
 
-PyTypeObject py_madfile_t = {PyObject_HEAD_INIT(&PyType_Type)0, "MadFile",
-                             sizeof(py_madfile),                0,
-                             /* standard methods */
-                             (destructor)py_madfile_dealloc,    (printfunc)0,
-                             (getattrfunc)py_madfile_getattr,   (setattrfunc)0,
-                             (cmpfunc)0,                        (reprfunc)0,
-                             /* type categories */
-                             0, /* as number */
-                             0, /* as sequence */
-                             0, /* as mapping */
-                             0, /* hash */
-                             0, /* binary */
-                             0, /* repr */
-                             0, /* getattro */
-                             0, /* setattro */
-                             0, /* as buffer */
-                             0, /* tp_flags */
-                             NULL};
+static PyMethodDef madfile_methods[] = {
+    {"read", py_madfile_read, METH_VARARGS, ""},
+    {"layer", py_madfile_layer, METH_VARARGS, ""},
+    {"mode", py_madfile_mode, METH_VARARGS, ""},
+    {"samplerate", py_madfile_samplerate, METH_VARARGS, ""},
+    {"bitrate", py_madfile_bitrate, METH_VARARGS, ""},
+    {"emphasis", py_madfile_emphasis, METH_VARARGS, ""},
+    {"total_time", py_madfile_total_time, METH_VARARGS, ""},
+    {"current_time", py_madfile_current_time, METH_VARARGS, ""},
+    {"seek_time", py_madfile_seek_time, METH_VARARGS, ""},
+    {NULL, 0, 0, NULL}};
+
+PyTypeObject py_madfile_t = {
+    PyVarObject_HEAD_INIT(&PyType_Type, 0) "MadFile", /* tp_name */
+    sizeof(py_madfile),                               /* tp_basicsize */
+    0,                                                /* tp_itemsize */
+    (destructor)py_madfile_dealloc,                   /* tp_dealloc */
+    0,                                                /* (tp_print) */
+    0,                                                /* (tp_getattr) */
+    0,                                                /* (tp_setattr) */
+    0,               /* (tp_reserved) [was tp_compare] */
+    0,               /* tp_repr */
+    0,               /* tp_as_number */
+    0,               /* tp_as_sequence */
+    0,               /* tp_as_mapping */
+    0,               /* tp_hash */
+    0,               /* tp_call */
+    0,               /* tp_str */
+    0,               /* tp_getattro */
+    0,               /* tp_setattro */
+    0,               /* tp_as_buffer */
+    0,               /* tp_flags */
+    0,               /* tp_doc */
+    0,               /* tp_traverse */
+    0,               /* tp_clear */
+    0,               /* tp_richcompare */
+    0,               /* tp_weaklistoffset */
+    0,               /* tp_iter */
+    0,               /* tp_iternext */
+    madfile_methods, /* tp_methods */
+    0,               /* tp_members */
+    0,               /* tp_getset */
+    0,               /* tp_base */
+    0,               /* tp_dict */
+    0,               /* tp_descr_get */
+    0,               /* tp_descr_set */
+    0,               /* tp_dictoffset */
+    0,               /* tp_init */
+    0,               /* tp_alloc */
+    0,               /* tp_new */
+    0,               /* tp_free */
+    0,               /* tp_is_gc */
+    0,               /* tp_bases */
+    0,               /* tp_mro */
+    0,               /* tp_cache */
+    0,               /* tp_subclasses */
+    0,               /* tp_weaklist */
+    0,               /* tp_del */
+    0,               /* tp_version_tag */
+};
 
 /* functions */
 
 PyObject *py_madfile_new(PyObject *self, PyObject *args) {
   py_madfile *mf = NULL;
   int close_file = 0;
+#if PY_MAJOR_VERSION >= 3
+  int fd;
+#endif
   char *fname;
   PyObject *fobject = NULL;
   char *initial;
   long ibytes = 0;
-  unsigned long int bufsiz = MAD_BUF_SIZE;
+  unsigned long int bufsize = MAD_BUF_SIZE;
   int n;
 
-  if (PyArg_ParseTuple(args, "s|l:MadFile", &fname, &bufsiz)) {
+  if (PyArg_ParseTuple(args, "s|l:MadFile", &fname, &bufsize)) {
+#if PY_MAJOR_VERSION < 3
     fobject = PyFile_FromString(fname, "r");
+#else
+    if ((fd = open(fname, O_RDONLY)) < 0) {
+      return NULL;
+    }
+    fobject = PyFile_FromFd(fd, fname, "r", -1, NULL, NULL, NULL, 1);
+#endif
     close_file = 1;
     if (fobject == NULL) {
       return NULL;
@@ -121,11 +188,11 @@ PyObject *py_madfile_new(PyObject *self, PyObject *args) {
   } else
     return NULL;
 
-  /* bufsiz must be an integer multiple of 4 */
-  if ((n = bufsiz % 4))
-    bufsiz -= n;
-  if (bufsiz <= 4096)
-    bufsiz = 4096;
+  /* bufsize must be an integer multiple of 4 */
+  if ((n = bufsize % 4))
+    bufsize -= n;
+  if (bufsize <= 4096)
+    bufsize = 4096;
 
   mf = PyObject_NEW(py_madfile, &py_madfile_t);
   Py_INCREF(fobject);
@@ -133,15 +200,15 @@ PyObject *py_madfile_new(PyObject *self, PyObject *args) {
   mf->close_file = close_file;
 
   /* initialise the mad structs */
-  mad_stream_init(&MAD_STREAM(mf));
-  mad_frame_init(&MAD_FRAME(mf));
-  mad_synth_init(&MAD_SYNTH(mf));
-  mad_timer_reset(&MAD_TIMER(mf));
+  mad_stream_init(&PYMAD_STREAM(mf));
+  mad_frame_init(&PYMAD_FRAME(mf));
+  mad_synth_init(&PYMAD_SYNTH(mf));
+  mad_timer_reset(&PYMAD_TIMER(mf));
 
   mf->framecount = 0;
 
-  mf->buffy = malloc(bufsiz * sizeof(unsigned char));
-  mf->bufsiz = bufsiz;
+  mf->input_buffer = malloc(bufsize * sizeof(unsigned char));
+  mf->bufsize = bufsize;
 
   /* explicitly call read to fill the buffer and have the frame header
    * data immediately available to the caller */
@@ -156,13 +223,13 @@ static void py_madfile_dealloc(PyObject *self, PyObject *args) {
   PyObject *o;
 
   if (PY_MADFILE(self)->fobject) {
-    mad_synth_finish(&MAD_SYNTH(self));
-    mad_frame_finish(&MAD_FRAME(self));
-    mad_stream_finish(&MAD_STREAM(self));
+    mad_synth_finish(&PYMAD_SYNTH(self));
+    mad_frame_finish(&PYMAD_FRAME(self));
+    mad_stream_finish(&PYMAD_STREAM(self));
 
-    free(MAD_BUFFY(self));
-    MAD_BUFFY(self) = NULL;
-    MAD_BUFSIZ(self) = 0;
+    free(PYMAD_BUFFER(self));
+    PYMAD_BUFFER(self) = NULL;
+    PYMAD_BUFSIZE(self) = 0;
 
     if (PY_MADFILE(self)->close_file) {
       o = PyObject_CallMethod(PY_MADFILE(self)->fobject, "close", NULL);
@@ -188,10 +255,10 @@ static unsigned long calc_total_time(PyObject *self) {
   int fnum;
 
   xing_init(&xing);
-  xing_parse(&xing, MAD_STREAM(self).anc_ptr, MAD_STREAM(self).anc_bitlen);
+  xing_parse(&xing, PYMAD_STREAM(self).anc_ptr, PYMAD_STREAM(self).anc_bitlen);
 
   if (xing.flags & XING_FRAMES) {
-    timer = MAD_FRAME(self).header.duration;
+    timer = PYMAD_FRAME(self).header.duration;
     mad_timer_multiply(&timer, xing.frames);
     r = mad_timer_count(timer, MAD_UNITS_MILLISECONDS);
   } else {
@@ -242,8 +309,8 @@ static unsigned long calc_total_time(PyObject *self) {
   return r;
 }
 
-/* convert the mad format to an unsigned short */
-static signed short int madfixed_to_short(mad_fixed_t sample) {
+/* convert the MAD fixed point format to a signed 16 bit int */
+static int16_t madfixed_to_int16(mad_fixed_t sample) {
   /* A fixed point number is formed of the following bit pattern:
    *
    * SWWWFFFFFFFFFFFFFFFFFFFFFFFFFFFF
@@ -257,13 +324,13 @@ static signed short int madfixed_to_short(mad_fixed_t sample) {
    * number.  It is not guaranteed to be constant over the different
    * platforms supported by libmad.
    *
-   * The unsigned short value is formed by the least significant
+   * The int16_t value is formed by the least significant
    * whole part bit, followed by the 15 most significant fractional
    * part bits.
    *
-   * This algorithm was taken from input/mad/mad_engine.c in
-   * alsaplayer, because the one in madlld sucked arse.
-   * http://www.geocrawler.com/archives/3/15440/2001/9/0/6629549/
+   * This algorithm was taken from input/mad/mad_engine.c in alsaplayer,
+   * which scales and rounds samples to 16 bits, unlike the version in
+   * madlld.
    */
 
   /* round */
@@ -280,8 +347,8 @@ static signed short int madfixed_to_short(mad_fixed_t sample) {
 }
 
 static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
-  PyObject *pybuf; /* return object containing output buffer*/
-  char *buffy;     /* output buffer */
+  PyObject *pybuf;        /* return object containing output buffer*/
+  int16_t *output = NULL; /* output buffer */
   unsigned int i;
   Py_ssize_t size;
   int nextframe = 0;
@@ -304,8 +371,8 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
 
     /* The input bucket must be filled if it becomes empty or if
      * it's the first execution on this file */
-    if ((MAD_STREAM(self).buffer == NULL) ||
-        (MAD_STREAM(self).error == MAD_ERROR_BUFLEN)) {
+    if ((PYMAD_STREAM(self).buffer == NULL) ||
+        (PYMAD_STREAM(self).error == MAD_ERROR_BUFLEN)) {
       Py_ssize_t readsize, remaining;
       unsigned char *readstart;
       PyObject *o_read;
@@ -328,13 +395,14 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
        *     frame at the highest observable bitrate (currently
        *     448kbps).
        */
-      if (MAD_STREAM(self).next_frame != NULL) {
-        remaining = MAD_STREAM(self).bufend - MAD_STREAM(self).next_frame;
-        memmove(MAD_BUFFY(self), MAD_STREAM(self).next_frame, remaining);
-        readstart = MAD_BUFFY(self) + remaining;
-        readsize = MAD_BUFSIZ(self) - remaining;
+      if (PYMAD_STREAM(self).next_frame != NULL) {
+        remaining = PYMAD_STREAM(self).bufend - PYMAD_STREAM(self).next_frame;
+        memmove(PYMAD_BUFFER(self), PYMAD_STREAM(self).next_frame, remaining);
+        readstart = PYMAD_BUFFER(self) + remaining;
+        readsize = PYMAD_BUFSIZE(self) - remaining;
       } else
-        readstart = MAD_BUFFY(self), readsize = MAD_BUFSIZ(self), remaining = 0;
+        readstart = PYMAD_BUFFER(self), readsize = PYMAD_BUFSIZE(self),
+        remaining = 0;
 
       /* Fill in the buffer.  If an error occurs, make like a tree */
       o_read =
@@ -344,7 +412,7 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
         Py_INCREF(Py_None);
         return Py_None;
       }
-      PyString_AsStringAndSize(o_read, &o_buffer, &readsize);
+      PyBytes_AsStringAndSize(o_read, &o_buffer, &readsize);
       // EOF?
       if (readsize == 0) {
         Py_DECREF(o_read);
@@ -356,9 +424,9 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
 
       /* Pipe the new buffer content to libmad's stream decode
        * facility */
-      mad_stream_buffer(&MAD_STREAM(self), MAD_BUFFY(self),
+      mad_stream_buffer(&PYMAD_STREAM(self), PYMAD_BUFFER(self),
                         readsize + remaining);
-      MAD_STREAM(self).error = 0;
+      PYMAD_STREAM(self).error = 0;
     }
 
     /* Decode the next mpeg frame.  The streams are read from the
@@ -390,26 +458,26 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
      * skip the faulty part and resync to the next frame.
      */
     Py_BEGIN_ALLOW_THREADS;
-    result = mad_frame_decode(&MAD_FRAME(self), &MAD_STREAM(self));
+    result = mad_frame_decode(&PYMAD_FRAME(self), &PYMAD_STREAM(self));
     Py_END_ALLOW_THREADS;
     if (result) {
-      if (MAD_RECOVERABLE(MAD_STREAM(self).error)) {
+      if (MAD_RECOVERABLE(PYMAD_STREAM(self).error)) {
         /* FIXME: prefer to return an error string to the caller
          * rather than print to stderr
         fprintf(stderr, "mad: recoverable frame level error: %s\n",
-                mad_stream_errorstr(&MAD_STREAM(self)));
+                mad_stream_errorstr(&PYMAD_STREAM(self)));
         fflush(stderr);
         */
         /* go onto the next frame */
         nextframe = 1;
       } else {
-        if (MAD_STREAM(self).error == MAD_ERROR_BUFLEN) {
+        if (PYMAD_STREAM(self).error == MAD_ERROR_BUFLEN) {
           /* not enough data to decode */
           nextframe = 1;
         } else {
           snprintf(errmsg, ERROR_MSG_SIZE,
                    "unrecoverable frame level error: %s",
-                   mad_stream_errorstr(&MAD_STREAM(self)));
+                   mad_stream_errorstr(&PYMAD_STREAM(self)));
           PyErr_SetString(PyExc_RuntimeError, errmsg);
           return NULL;
         }
@@ -429,58 +497,53 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
    * of mad's timer module receive their mad_timer_t arguments by
    * value! */
   PY_MADFILE(self)->framecount++;
-  mad_timer_add(&MAD_TIMER(self), MAD_FRAME(self).header.duration);
+  mad_timer_add(&PYMAD_TIMER(self), PYMAD_FRAME(self).header.duration);
 
   /* Once decoded, the frame can be synthesised to PCM samples.
    * No errors are reported by mad_synth_frame() */
-  mad_synth_frame(&MAD_SYNTH(self), &MAD_FRAME(self));
+  mad_synth_frame(&PYMAD_SYNTH(self), &PYMAD_FRAME(self));
 
   Py_END_ALLOW_THREADS;
 
   /* Create the buffer to store the PCM samples in, so python can
-   * use it.  We do 4 pointer increments per sample in the buffer,
-   * so make it 4 times as big as the number of samples */
-  size = MAD_SYNTH(self).pcm.length * 4;
-  pybuf = PyBuffer_New(size);
-  PyObject_AsWriteBuffer(pybuf, (void *)&buffy, &size);
+   * use it.  We do 2 pointer increments per sample in the buffer,
+   * so make it 2 times as big as the number of samples */
+  size = PYMAD_SYNTH(self).pcm.length * 2 * sizeof(int16_t);
+
+  output = malloc(size);
+  if (!output) {
+    PyErr_SetString(PyExc_MemoryError,
+                    "could not allocate memory for output buffer");
+    return NULL;
+  }
+  pybuf = PyByteArray_FromStringAndSize((const char *)output, size);
 
   /* die if we don't have the space */
-  if (size < MAD_SYNTH(self).pcm.length * 4) {
+  if (size < PYMAD_SYNTH(self).pcm.length * 4) {
     PyErr_SetString(PyExc_MemoryError, "allocated buffer too small");
     return NULL;
   }
 
   Py_BEGIN_ALLOW_THREADS;
 
-  /* Synthesised samples must be converted from mad's fixed point
-   * format to the consumer format.  Here we use signed 16 bit
-   * big endian ints on two channels.  Integer samples are
-   * temporarily stored in a buffer that is flushed when full. */
-  for (i = 0; i < MAD_SYNTH(self).pcm.length; i++) {
-    signed short sample;
+  /* Synthesised samples must be converted from mad's fixed point format to the
+   * consumer format -- use signed 16 bit big-endian ints on two channels.
+   * Integer samples are temporarily stored in a buffer that is flushed when
+   * full. */
+  for (i = 0; i < PYMAD_SYNTH(self).pcm.length; i++) {
+    int16_t sample;
 
     /* left channel */
-    sample = madfixed_to_short(MAD_SYNTH(self).pcm.samples[0][i]);
-#ifdef BIGENDIAN
-    *(buffy++) = sample >> 8;
-    *(buffy++) = sample & 0xFF;
-#else
-    *(buffy++) = sample & 0xFF;
-    *(buffy++) = sample >> 8;
-#endif
+    *(output++) = sample =
+        madfixed_to_int16(PYMAD_SYNTH(self).pcm.samples[0][i]);
 
     /* right channel.
      * if the decoded stream is monophonic then the right channel
      * is the same as the left one */
-    if (MAD_NCHANNELS(&MAD_FRAME(self).header) == 2)
-      sample = madfixed_to_short(MAD_SYNTH(self).pcm.samples[1][i]);
-#ifdef Bigendian
-    *(buffy++) = sample >> 8;
-    *(buffy++) = sample & 0xFF;
-#else
-    *(buffy++) = sample & 0xFF;
-    *(buffy++) = sample >> 8;
-#endif
+    if (MAD_NCHANNELS(&PYMAD_FRAME(self).header) == 2)
+      sample = madfixed_to_int16(PYMAD_SYNTH(self).pcm.samples[1][i]);
+
+    *(output++) = sample;
   }
 
   Py_END_ALLOW_THREADS;
@@ -490,27 +553,27 @@ static PyObject *py_madfile_read(PyObject *self, PyObject *args) {
 
 /* return the MPEG layer */
 static PyObject *py_madfile_layer(PyObject *self, PyObject *args) {
-  return PyInt_FromLong(MAD_FRAME(self).header.layer);
+  return PyInt_FromLong(PYMAD_FRAME(self).header.layer);
 }
 
 /* return the channel mode */
 static PyObject *py_madfile_mode(PyObject *self, PyObject *args) {
-  return PyInt_FromLong(MAD_FRAME(self).header.mode);
+  return PyInt_FromLong(PYMAD_FRAME(self).header.mode);
 }
 
 /* return the stream samplerate */
 static PyObject *py_madfile_samplerate(PyObject *self, PyObject *args) {
-  return PyInt_FromLong(MAD_FRAME(self).header.samplerate);
+  return PyInt_FromLong(PYMAD_FRAME(self).header.samplerate);
 }
 
 /* return the stream bitrate */
 static PyObject *py_madfile_bitrate(PyObject *self, PyObject *args) {
-  return PyInt_FromLong(MAD_FRAME(self).header.bitrate);
+  return PyInt_FromLong(PYMAD_FRAME(self).header.bitrate);
 }
 
 /* return the emphasis value */
 static PyObject *py_madfile_emphasis(PyObject *self, PyObject *args) {
-  return PyInt_FromLong(MAD_FRAME(self).header.emphasis);
+  return PyInt_FromLong(PYMAD_FRAME(self).header.emphasis);
 }
 
 /* return the estimated playtime of the track, in milliseconds */
@@ -521,7 +584,7 @@ static PyObject *py_madfile_total_time(PyObject *self, PyObject *args) {
 /* return the current position in the track, in milliseconds */
 static PyObject *py_madfile_current_time(PyObject *self, PyObject *args) {
   return PyInt_FromLong(
-      mad_timer_count(MAD_TIMER(self), MAD_UNITS_MILLISECONDS));
+      mad_timer_count(PYMAD_TIMER(self), MAD_UNITS_MILLISECONDS));
 }
 
 /* seek playback to the given position, in milliseconds, from the start
@@ -562,29 +625,11 @@ static PyObject *py_madfile_seek_time(PyObject *self, PyObject *args) {
   }
   Py_DECREF(o);
 
-  mad_stream_init(&MAD_STREAM(self));
-  mad_frame_init(&MAD_FRAME(self));
-  mad_synth_init(&MAD_SYNTH(self));
-  mad_timer_reset(&MAD_TIMER(self));
-  mad_timer_set(&MAD_TIMER(self), 0, pos, 1000);
+  mad_stream_init(&PYMAD_STREAM(self));
+  mad_frame_init(&PYMAD_FRAME(self));
+  mad_synth_init(&PYMAD_SYNTH(self));
+  mad_timer_reset(&PYMAD_TIMER(self));
+  mad_timer_set(&PYMAD_TIMER(self), 0, pos, 1000);
 
   return Py_None;
 }
-
-/* housekeeping */
-
-static PyMethodDef madfile_methods[] = {
-    {"read", py_madfile_read, METH_VARARGS, ""},
-    {"layer", py_madfile_layer, METH_VARARGS, ""},
-    {"mode", py_madfile_mode, METH_VARARGS, ""},
-    {"samplerate", py_madfile_samplerate, METH_VARARGS, ""},
-    {"bitrate", py_madfile_bitrate, METH_VARARGS, ""},
-    {"emphasis", py_madfile_emphasis, METH_VARARGS, ""},
-    {"total_time", py_madfile_total_time, METH_VARARGS, ""},
-    {"current_time", py_madfile_current_time, METH_VARARGS, ""},
-    {"seek_time", py_madfile_seek_time, METH_VARARGS, ""},
-    {NULL, 0, 0, NULL}};
-
-static PyObject *py_madfile_getattr(PyObject *self, char *name) {
-  return Py_FindMethod(madfile_methods, self, name);
-}
